"""
Setup:
Follow setup instruction from the github documentation to create personal access token
https://docs.github.com/en/authentication/keeping-your-account-and-data-secure/creating-a-personal-access-token
"""

from typing import Annotated
from httpx import Client

from admyral.action import action, ArgumentMetadata
from admyral.context import ctx
from admyral.typings import JsonValue
from admyral.utils.collections import is_not_empty


def get_github_enterprise_client(access_token: str, enterprise: str) -> Client:
    return Client(
        base_url=f"https://api.github.com/enterprises/{enterprise}",
        headers={
            "Authorization": f"Bearer {access_token}",
            "Accept": "application/vnd.github.v3+json",
        },
    )


@action(
    display_name="Search Enterprise Audit Logs",
    display_namespace="GitHub",
    description="Search GitHub audit logs for an enterprise.",
    secrets_placeholders=["GITHUB_ENTERPRISE_SECRET"],
)
def search_github_enterprise_audit_logs(
    enterprise: Annotated[
        str,
        ArgumentMetadata(
            display_name="Enterprise",
            description="The slug version of the enterprise name. You can also substitute this value with the enterprise id.",
        ),
    ],
    filter: Annotated[
        str,
        ArgumentMetadata(
            display_name="Filter",
            description="Filter to apply to the query.",
        ),
    ] = None,
    start_time: Annotated[
        str | None,
        ArgumentMetadata(
            display_name="Start Time",
            description="The start time for the cases to list. Must be in ISO 8601 format (YYYY-MM-DDTHH:MM:SSZ).",
        ),
    ] = None,
    end_time: Annotated[
        str | None,
        ArgumentMetadata(
            display_name="End Time",
            description="The end time for the cases to list. Must be in ISO 8601 format (YYYY-MM-DDTHH:MM:SSZ).",
        ),
    ] = None,
    limit: Annotated[
        int | None,
        ArgumentMetadata(
            display_name="Limit",
            description="Maximum number of logs to return.",
        ),
    ] = None,
) -> list[dict[str, JsonValue]]:
    # https://docs.github.com/en/enterprise-cloud@latest/rest/enterprise-admin/audit-log?apiVersion=2022-11-28#get-the-audit-log-for-an-enterprise
    # https://docs.github.com/en/search-github/getting-started-with-searching-on-github/understanding-the-search-syntax#query-for-dates

    secret = ctx.get().secrets.get("GITHUB_ENTERPRISE_SECRET")
    access_token = secret["access_token"]

    with get_github_enterprise_client(access_token, enterprise) as client:
        params = {"order": "asc", "per_page": 100}

        phrases = []

        if filter:
            phrases.append(filter)

        if start_time and end_time:
            phrases.append(f"created:{start_time}..{end_time}")
        elif start_time:
            phrases.append(f"created:>={start_time}")
        elif end_time:
            phrases.append(f"created:<={end_time}")

        if is_not_empty(phrases):
            params["phrase"] = " ".join(phrases)

        url = "/audit-log"
        events = []
        while limit is None or len(events) < limit:
            response = client.get(
                url,
                params=params,
            )
            response.raise_for_status()
            events.extend(response.json())

            if "next" in response.links:
                url = response.links["next"]["url"][len(str(client.base_url)) :]
                params = None
            else:
                break

        return events if limit is None else events[:limit]



from typing import Any


class Singleton(type):
    """
    Metaclass for Singleton.
    Source: https://stackoverflow.com/a/33201
    """

    def __init__(cls, *args: Any, **kwargs: Any) -> None:
        super(Singleton, cls).__init__(*args, **kwargs)
        cls._instance = None

    def __call__(cls, *args: Any, **kwds: Any) -> Any:
        if cls._instance is None:
            cls._instance = super(Singleton, cls).__call__(*args, **kwds)
        return cls._instance



from pydantic import BaseModel

from admyral.typings import JsonValue


# class ArgumentType(str, Enum):
#     STRING = "string"
"""
Types:

- str
- int, float
- JsonValue
- list
- dict
- None



- required => some type does not contain None
    - can have a default value
- optional => can be None (where None must be set explicitly)
    - default is None or something else

"""


class Argument(BaseModel):
    arg_name: str
    display_name: str
    description: str
    arg_type: str  # TODO: make this an enum
    is_optional: bool
    default_value: JsonValue | None


class PythonAction(BaseModel):
    """
    Model for a custom Python action.
    """

    action_type: str
    import_statements: str
    code: str
    arguments: list[Argument]

    # decorator arguments
    display_name: str
    display_namespace: str
    description: str | None = None
    secrets_placeholders: list[str] | None = None
    requirements: list[str] | None = None


class ActionMetadata(BaseModel):
    action_type: str
    display_name: str
    display_namespace: str
    description: str | None = None
    secrets_placeholders: list[str] = []
    arguments: list[Argument] = []



from typing import Optional
from fastapi import APIRouter, status

from admyral.models import PythonAction
from admyral.server.deps import get_admyral_store


router = APIRouter()


@router.post("/push", status_code=status.HTTP_204_NO_CONTENT)
async def push_action(python_action: PythonAction):
    """
    Push a Python action to the store. If the action for the provided action type already exists for the user,
    it will be overwritten.

    Args:
        python_action: The Python action object.
    """
    await get_admyral_store().store_action(python_action)


@router.get(
    "/{action_type}", status_code=status.HTTP_200_OK, response_model=PythonAction
)
async def get_action(action_type: str) -> Optional[PythonAction]:
    """
    Get a Python action by its action type.

    Args:
        action_type: The action type.

    Returns:
        The Python action object.
    """
    return await get_admyral_store().get_action(action_type)



import os
from typing import Annotated, Literal
from openai import OpenAI

from admyral.action import action, ArgumentMetadata


@action(
    display_name="AI Action",
    display_namespace="Admyral",
    description="Use advanced AI models to perform complex tasks, such as categorization, "
    "analysis, summarization, or decision support.",
)
def ai_action(
    model: Annotated[
        Literal["gpt-4", "gpt-4o", "gpt-4-turbo"],
        ArgumentMetadata(
            display_name="Model",
            description="The model to use for the AI action",
        ),
    ],
    prompt: Annotated[
        str,
        ArgumentMetadata(
            display_name="Prompt",
            description="The prompt to use for the AI action",
        ),
    ],
    top_p: Annotated[
        float | None,
        ArgumentMetadata(
            display_name="Top P",
            description="Value between 0 and 1. An alternative to sampling with temperature, "
            "called nucleus sampling, where the model considers the results of the tokens "
            "with Top P probability mass. So 0.1 means only the tokens comprising the top 10% "
            "probability mass are considered. We generally recommend altering this or temperature "
            "but not both.",
        ),
    ] = None,
    temperature: Annotated[
        float | None,
        ArgumentMetadata(
            display_name="Temperature",
            description="What sampling temperature to use, between 0 and 2. Higher values like 0.8 "
            "will make the output more random, while lower values like 0.2 will make it more focused "
            "and deterministic. We generally recommend altering this or Top P but not both.",
        ),
    ] = None,
) -> str:
    api_key = os.environ.get("OPENAI_API_KEY")
    if not api_key:
        raise ValueError(
            "OPENAI_API_KEY environment variable is not set. Please set it to use the send_email action."
        )
    client = OpenAI(api_key=api_key)
    chat_completion = client.chat.completions.create(
        messages=[{"role": "user", "content": prompt}],
        model=model,
        top_p=top_p,
        temperature=temperature,
    )
    return chat_completion.choices[0].message.content



from admyral.actions.integrations.compliance.retool import (
    list_retool_inactive_users,
    list_groups_per_user,
)
from admyral.actions.integrations.compliance.one_password import (
    list_1password_audit_events,
)
from admyral.actions.integrations.compliance.github import (
    search_github_enterprise_audit_logs,
)

__all__ = [
    "list_retool_inactive_users",
    "list_groups_per_user",
    "list_1password_audit_events",
    "search_github_enterprise_audit_logs",
]



import click

from admyral.client import AdmyralClient


@click.group()
@click.pass_context
def cli(ctx):
    """Admyral CLI

    Admyral is an open-source Python SDK enabling Security Engineers to build complex workflow automations using Python — blazingly fast.
    """
    ctx.obj = AdmyralClient()



from typing import Annotated
from anthropic import Anthropic
from anthropic.types import TextBlock

from admyral.action import action, ArgumentMetadata
from admyral.context import ctx


@action(
    display_name="Chat Completion",
    display_namespace="Anthropic",
    description="Use advanced AI models from Azure OpenAI to perform complex tasks, such as categorization, "
    "analysis, summarization, or decision support.",
    secrets_placeholders=["ANTHROPIC_SECRET"],
)
def anthropic_chat_completion(
    model: Annotated[
        str,
        ArgumentMetadata(
            display_name="Model",
            description="The model to use.",
        ),
    ],
    prompt: Annotated[
        str,
        ArgumentMetadata(
            display_name="Prompt",
            description="The prompt to use for the chat completion.",
        ),
    ],
    top_p: Annotated[
        float | None,
        ArgumentMetadata(
            display_name="Top P",
            description="Value between 0 and 1. An alternative to sampling with temperature, "
            "called nucleus sampling, where the model considers the results of the tokens "
            "with Top P probability mass. So 0.1 means only the tokens comprising the top 10% "
            "probability mass are considered. We generally recommend altering this or temperature "
            "but not both.",
        ),
    ] = None,
    temperature: Annotated[
        float | None,
        ArgumentMetadata(
            display_name="Temperature",
            description="What sampling temperature to use, between 0 and 2. Higher values like 0.8 "
            "will make the output more random, while lower values like 0.2 will make it more focused "
            "and deterministic. We generally recommend altering this or Top P but not both.",
        ),
    ] = None,
    max_tokens: Annotated[
        int | None,
        ArgumentMetadata(
            display_name="Max Tokens",
            description="The maximum number of tokens to generate in the completion.",
        ),
    ] = None,
    stop_tokens: Annotated[
        list[str] | None,
        ArgumentMetadata(
            display_name="Stop Tokens",
            description="A list of tokens at which to stop the completion.",
        ),
    ] = None,
) -> str:
    # TODO: error handling
    secret = ctx.get().secrets.get("ANTHROPIC_SECRET")
    api_key = secret["api_key"]

    client = Anthropic(api_key=api_key)

    message = client.messages.create(
        messages=[
            {
                "role": "user",
                "content": prompt,
            }
        ],
        model=model,
        top_p=top_p,
        temperature=temperature,
        max_tokens=max_tokens,
        stop_sequences=stop_tokens,
    )

    # TODO: what is the best behavior for returning the content blocks?
    return "\n".join(
        [
            block.text
            if isinstance(block, TextBlock)
            else f"Tool(id={block.id}, input={block.input}, name={block.name})"
            for block in message.content
        ]
    )



import asyncio

from admyral.logger import get_logger
from admyral.server.deps import get_admyral_store
from admyral.config.config import GlobalConfig


logger = get_logger(__name__)


async def cleanup_pip_lockfile_cache(cleanup_interval: int):
    while True:
        await asyncio.sleep(cleanup_interval)
        logger.info("Cleaning up pip lockfile cache...")
        await get_admyral_store().delete_expired_cached_pip_lockfile()
        logger.info("Finished cleaning up pip lockfile cache.")


def start_background_tasks():
    logger.info("Starting background tasks...")

    asyncio.create_task(
        cleanup_pip_lockfile_cache(GlobalConfig().pip_lockfile_cache_cleanup_interval)
    )
    logger.info("Started pip lockfile cache cleanup background task.")



from typing import Any
from temporalio import activity
from uuid import uuid4
import time

from admyral.workers.shared_worker_state import SharedWorkerState
from admyral.logger import get_logger


logger = get_logger(__name__)


@activity.defn
async def init_workflow_run(
    workflow_id: str,
    source_name: str,
    payload: dict[str, Any],
) -> str:
    start = time.monotonic_ns()
    run_id = str(uuid4())
    step_id = str(uuid4())
    await SharedWorkerState.get_store().init_workflow_run(
        run_id,
        step_id,
        workflow_id,
        source_name,
        payload,
    )
    end = time.monotonic_ns()
    logger.info(f"Initialized workflow run {run_id} in {(end - start) / 1_000_000}ms")
    return run_id



from typing import Annotated, Literal
import base64
from httpx import Client

from admyral.typings import JsonValue
from admyral.action import action, ArgumentMetadata
from admyral.context import ctx
from admyral.utils.collections import is_empty


def get_jira_client(domain: str, email: str, api_key: str) -> Client:
    api_key_base64 = base64.b64encode(f"{email}:{api_key}".encode()).decode()
    return Client(
        base_url=f"https://{domain}/rest/api/3",
        headers={
            "Authorization": f"Basic {api_key_base64}",
            "Content-Type": "application/json",
            "Accept": "application/json",
        },
    )


@action(
    display_name="Create Jira Issue",
    display_namespace="Jira",
    description="Create a Jira issue",
    secrets_placeholders=["JIRA_SECRET"],
)
def create_jira_issue(
    summary: Annotated[
        str,
        ArgumentMetadata(
            display_name="Summary",
            description="The summary of the issue",
        ),
    ],
    project_id: Annotated[
        str,
        ArgumentMetadata(
            display_name="Project ID",
            description="The ID of the project to create the issue in",
        ),
    ],
    issue_type: Annotated[
        str,
        ArgumentMetadata(
            display_name="Issue Type",
            description="The issue type",
        ),
    ],
    description: Annotated[
        JsonValue | None,
        ArgumentMetadata(
            display_name="Description",
            description="The description of the issue in Atlassian Document Format",
        ),
    ] = None,
    assignee: Annotated[
        str | None,
        ArgumentMetadata(
            display_name="Assignee",
            description="The ID of the assignee",
        ),
    ] = None,
    labels: Annotated[
        list[str],
        ArgumentMetadata(
            display_name="Labels",
            description="The labels to add to the issue",
        ),
    ] = [],
    priority: Annotated[
        Literal["Lowest", "Low", "Medium", "High", "Highest"] | None,
        ArgumentMetadata(
            display_name="Priority",
            description="The priority of the issue",
        ),
    ] = None,
    custom_fields: Annotated[
        dict[str, JsonValue],
        ArgumentMetadata(
            display_name="Custom Fields",
            description="The custom fields to add to the issue as a JSON object",
        ),
    ] = {},
) -> JsonValue:
    """
    How to find the project id?

        curl --request GET \
            --url https://christesting123.atlassian.net/rest/api/latest/project  \
            --header 'Authorization: Basic <base64-encoded-email:api_key-pair>'


    How to find the user id?

        curl --request GET \
            --url https://<your_domain>.atlassian.net/rest/api/3/myself  \
            --header 'Authorization: Basic <base64-encoded-email:api_key-pair>'
        
        or

        go to https://<your_domain>.atlassian.net/people
    
    How to find the issue type id?

        curl --request GET \
            --url https://<your_domain>.atlassian.net/rest/api/3/issuetype  \
            --header 'Authorization: Basic <base64-encoded-email:api_key-pair>'

    """
    # Atlassian Document Format: https://developer.atlassian.com/cloud/jira/platform/apis/document/structure/
    # https://developer.atlassian.com/cloud/jira/platform/rest/v3/api-group-issues/#api-rest-api-3-issue-post
    secret = ctx.get().secrets.get("JIRA_SECRET")
    domain = secret["domain"]
    email = secret["email"]
    api_key = secret["api_key"]

    body = {
        "fields": {
            "project": {
                "id": project_id,
            },
            "summary": summary,
            "issuetype": {
                "name": issue_type,
            },
        }
    }

    if description:
        body["fields"]["description"] = description

    if assignee:
        body["fields"]["assignee"] = {
            "id": assignee,
        }

    if labels:
        body["fields"]["labels"] = labels

    if priority:
        body["fields"]["priority"] = {
            "name": priority,
        }

    for key, value in custom_fields.items():
        body["fields"][key] = value

    with get_jira_client(domain, email, api_key) as client:
        response = client.post(
            "/issue",
            json=body,
        )
        response.raise_for_status()
        return response.json()


@action(
    display_name="Update Jira Issue Status",
    display_namespace="Jira",
    description="Update a Jira issue status",
    secrets_placeholders=["JIRA_SECRET"],
)
def update_jira_issue_status(
    issue_id_or_key: Annotated[
        str,
        ArgumentMetadata(
            display_name="Issue ID or Key",
            description="The ID or the key of the issue",
        ),
    ],
    transition_id: Annotated[
        str,
        ArgumentMetadata(
            display_name="Transition ID",
            description="The ID of the transition",
        ),
    ],
) -> None:
    # https://developer.atlassian.com/cloud/jira/platform/rest/v3/api-group-issues/#api-rest-api-3-issue-issueidorkey-transitions-post
    # https://community.atlassian.com/t5/Jira-questions/How-do-i-find-a-transition-ID/qaq-p/2113213
    secret = ctx.get().secrets.get("JIRA_SECRET")
    domain = secret["domain"]
    email = secret["email"]
    api_key = secret["api_key"]

    with get_jira_client(domain, email, api_key) as client:
        response = client.post(
            f"/issue/{issue_id_or_key}/transitions",
            json={
                "transition": {
                    "id": transition_id,
                }
            },
        )
        response.raise_for_status()


@action(
    display_name="Comment Jira Issue",
    display_namespace="Jira",
    description="Comment Jira issue",
    secrets_placeholders=["JIRA_SECRET"],
)
def comment_jira_issue_status(
    issue_id_or_key: Annotated[
        str,
        ArgumentMetadata(
            display_name="Issue ID or Key",
            description="The ID or the key of the issue",
        ),
    ],
    comment: Annotated[
        JsonValue | None,
        ArgumentMetadata(
            display_name="Comment",
            description="The comment in Atlassian Document Format",
        ),
    ] = None,
) -> None:
    # https://developer.atlassian.com/cloud/jira/platform/rest/v3/api-group-issues/#api-rest-api-3-issue-issueidorkey-transitions-post
    secret = ctx.get().secrets.get("JIRA_SECRET")
    domain = secret["domain"]
    email = secret["email"]
    api_key = secret["api_key"]

    with get_jira_client(domain, email, api_key) as client:
        response = client.post(
            f"/issue/{issue_id_or_key}/comment",
            json={"body": comment},
        )
        response.raise_for_status()


@action(
    display_name="Jira Issue Search",
    display_namespace="Jira",
    description="Search Jira issues using JQL",
    secrets_placeholders=["JIRA_SECRET"],
)
def search_jira_issues(
    jql: Annotated[
        str,
        ArgumentMetadata(
            display_name="JQL",
            description="The JQL query to search for issues",
        ),
    ],
    limit: Annotated[
        int | None,
        ArgumentMetadata(
            display_name="Limit",
            description="The maximum number of issues to return",
        ),
    ] = 1000,
) -> list[JsonValue]:
    # https://developer.atlassian.com/cloud/jira/platform/rest/v3/api-group-issue-search/#api-rest-api-3-search-get
    secret = ctx.get().secrets.get("JIRA_SECRET")
    domain = secret["domain"]
    email = secret["email"]
    api_key = secret["api_key"]

    with get_jira_client(domain, email, api_key) as client:
        offset = 0
        issues = []

        while limit is None or len(issues) < limit:
            response = client.get(
                "/search",
                params={"jql": jql, "maxResults": 100, "startAt": offset},
            )
            response.raise_for_status()
            data = response.json()

            issues.extend(data["issues"])
            offset = len(issues)
            if offset == data["total"]:
                break

        return issues if limit is None else issues[:limit]


@action(
    display_name="Get Audit Records",
    display_namespace="Jira",
    description="Get audit records",
    secrets_placeholders=["JIRA_SECRET"],
)
def get_jira_audit_records(
    start_date: Annotated[
        str | None,
        ArgumentMetadata(
            display_name="Start Date",
            description="The start date in in ISO-8601 format (YYYY-MM-DDTHH:MM:SSZ).",
        ),
    ] = None,
    end_date: Annotated[
        str | None,
        ArgumentMetadata(
            display_name="End Date",
            description="The end date in in ISO-8601 format (YYYY-MM-DDTHH:MM:SSZ).",
        ),
    ] = None,
    filter: Annotated[
        list[str] | None,
        ArgumentMetadata(
            display_name="Filter",
            description="A list of strings to match with audit field content. The strings must not contain spaces.",
        ),
    ] = None,
    limit: Annotated[
        int | None,
        ArgumentMetadata(
            display_name="Limit",
            description="The maximum number of audit records to return.",
        ),
    ] = None,
):
    # https://developer.atlassian.com/cloud/jira/platform/rest/v3/api-group-audit-records/#api-rest-api-3-auditing-record-get
    secret = ctx.get().secrets.get("JIRA_SECRET")
    domain = secret["domain"]
    email = secret["email"]
    api_key = secret["api_key"]

    with get_jira_client(domain, email, api_key) as client:
        offset = 0
        logs = []

        params = {}
        if start_date is not None:
            params["from"] = start_date
        if end_date is not None:
            params["to"] = end_date
        if filter is not None:
            if any(" " in f for f in filter):
                raise ValueError("Filter strings must not contain spaces.")
            params["filter"] = " ".join(filter)

        while limit is None or len(logs) < limit:
            params["offset"] = offset
            response = client.get("/auditing/record", params=params)
            response.raise_for_status()
            data = response.json()

            logs.extend(data["records"])
            offset = len(logs)
            if offset == data["total"] or is_empty(data["records"]):
                break

        return logs if limit is None else logs[:limit]



"""add completed_at/failed_at to WorkflowRun and error to WorkflowRunStep

Revision ID: 5ebc9fcfc1c7
Revises: 014da66fd11f
Create Date: 2024-09-05 19:49:37.980623

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
import sqlmodel  # noqa F401


# revision identifiers, used by Alembic.
revision: str = "5ebc9fcfc1c7"
down_revision: Union[str, None] = "014da66fd11f"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("workflow_run_steps", sa.Column("error", sa.TEXT(), nullable=True))
    op.add_column(
        "workflow_runs", sa.Column("completed_at", sa.TIMESTAMP(), nullable=True)
    )
    op.add_column(
        "workflow_runs", sa.Column("failed_at", sa.TIMESTAMP(), nullable=True)
    )

    # By default we mark all previous runs as successfully completed.
    op.execute("UPDATE workflow_runs SET completed_at = created_at")
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("workflow_runs", "failed_at")
    op.drop_column("workflow_runs", "completed_at")
    op.drop_column("workflow_run_steps", "error")
    # ### end Alembic commands ###



import json
import re

from admyral.typings import JsonValue


REFERENCE_REGEX = re.compile(r"{{((?!}}).)*}}")


"""

Behavior:

- If something is wrapped in quotes, it is a string.
    - Special behavior: If we have something like `"some string"` as input, then the quotes will be kept
    - If we have something like `"123"` as input, then the quotes will be removed and we will simply store 123 as a string
- In all other cases, we will JSON-deserialize the string and store the result.

    

Types:

- If a reference not wrapped in a string is used, it simply returns the value it references.
    Example 1: `{{ a['b'] }}` with `a['b'] = "abc"` will be resolved to `"abc"`.
    Example 2: `{{ a['b'] }}` with `a['b'] = 123` will be resolved to `123`.
- If a reference is wrapped in a string, it returns the reference as a string.
  - Note if it is a string within a string (e.g. `"... {{ a['b'] }} ..."`), it will not wrap the referenced value in quotes.
    Example: `"... {{ a['b'] }} ..."` with `a['b'] = "abc"` will be resolved to `"... abc ..."`.

"""


def _unescape_string(value: str) -> str:
    return value.encode("utf-8").decode("unicode_escape")


def _escape_string(value: str) -> str:
    return value.encode("unicode_escape").decode("utf-8")


def _is_float(value: str) -> bool:
    try:
        float(value)
        return True
    except ValueError:
        return False


def _is_string_escaped_json_value(value: str) -> bool:
    if not (
        _is_float(value)
        or value.isdecimal()
        or value in ("true", "false")
        or (value.startswith("{") and value.endswith("}"))
        or (value.startswith("[") and value.endswith("]"))
        or value == "null"
    ):
        return False
    try:
        json.loads(value)
        return True
    except json.JSONDecodeError:
        return False


def _handle_value_inside_container(value: JsonValue) -> str:
    if not isinstance(value, str):
        return serialize_json_with_reference(value)

    if value.startswith("{{") and value.endswith("}}"):
        return value
    return json.dumps(value)


def serialize_json_with_reference(value: JsonValue) -> str:
    if value is None or isinstance(value, (bool, int, float)):
        return json.dumps(value)

    if isinstance(value, str):
        # handle string escaping for ints, floats, bools, dicts, and lists
        if _is_string_escaped_json_value(value):
            return f'"{_escape_string(value)}"'
        return _escape_string(value)

    if isinstance(value, list):
        content = ", ".join(_handle_value_inside_container(item) for item in value)
        return f"[{content}]"

    if isinstance(value, dict):
        content = ", ".join(
            f"{_handle_value_inside_container(k)}: {_handle_value_inside_container(v)}"
            for k, v in value.items()
        )
        return f"{{{content}}}"

    raise ValueError(f"Unsupported value type: {type(value)}")


def deserialize_json_with_reference(value: str) -> JsonValue:
    value = value.strip().strip("\n")
    value = _unescape_string(value)

    if value == "":
        return None

    # handle string escaping for ints, floats, bools, dicts, and lists
    if value.startswith('"') and value.endswith('"'):
        return value[1:-1]

    # Handle pure reference
    if value.startswith("{{") and value.endswith("}}"):
        return value

    # Handle Integers
    if value.isdecimal():
        return int(value)

    # Handle Floats
    if _is_float(value):
        return float(value)

    # Handle Booleans
    if value in ("true", "false"):
        return value == "true"

    # Handle Null
    if value == "null":
        return None

    is_list = value.startswith("[") and value.endswith("]")
    is_dict = value.startswith("{") and value.endswith("}")
    if not is_dict and not is_list:
        # just a normal string
        return value

    # Handle Arrays and Objects
    # We might have something like the following string:
    # [{{ a['b'][0]["c"] }}, "something before and after", 42, true]
    #
    # => we must wrap the references within quotes such that we can do json.loads

    # Compute which positions are within quotes (i.e., "..." or '...')
    is_within_string = [False for _ in range(len(value))]
    idx = 0
    while idx < len(value):
        # Find the next " or '
        while idx < len(value) and value[idx] not in ('"', "'"):
            if value[idx] == "{" and idx + 1 < len(value) and value[idx + 1] == "{":
                # Skip reference: {{ <some-reference> }}
                idx += 2
                while idx < len(value) and (
                    value[idx] != "}"
                    or (idx + 1 < len(value) and value[idx + 1] != "}")
                ):
                    idx += 1
            idx += 1

        if idx == len(value):
            break

        # we have found a quote (" or ')
        expected_quote = value[idx]
        idx += 1
        while idx < len(value) and value[idx] != expected_quote:
            is_within_string[idx] = True

            if value[idx] == "{" and idx + 1 < len(value) and value[idx + 1] == "{":
                # Skip reference: {{ <some-reference> }}
                is_within_string[idx + 1] = True

                idx += 2
                while idx < len(value) and (
                    value[idx] != "}"
                    or (idx + 1 < len(value) and value[idx + 1] != "}")
                ):
                    is_within_string[idx] = True
                    idx += 1

            idx += 1

        idx += 1

    # Iterate over references and check whether we need to wrap them within "..."
    replacements = []
    for match in REFERENCE_REGEX.finditer(value):
        start, end = match.span()
        if is_within_string[start]:
            # Already within a quote, however, we must still replace "
            replacements.append(
                (start, end, match.group().replace('\\"', '"').replace('"', '\\"'))
            )
        else:
            # We need to wrap the reference within quotes because it is currently not
            # within quotes
            replacements.append(
                (
                    start,
                    end,
                    f'"{match.group().replace('\\"', '\"').replace('\"', '\\"')}"',
                )
            )

    # Wrap references into qutoes which are not yet within quotes
    # We can't simply just use regex replace because a reference might be used multiple times
    # where some usages are within quotes and some are not.
    out = [value]
    for start, end, replacement in reversed(replacements):
        str1 = out[-1][end:]
        str2 = out[-1][:start]

        out[-1] = str1
        out.append(replacement)
        out.append(str2)

    out = "".join(reversed(out))

    try:
        return json.loads(out)
    except json.JSONDecodeError:
        # normal string
        return value



from typing import Annotated
import json
import time

from admyral.action import action, ArgumentMetadata
from admyral.typings import JsonValue


@action(
    display_name="Deserialize JSON String",
    display_namespace="Admyral",
    description="Deserializes a JSON string.",
)
def deserialize_json_string(
    serialized_json: Annotated[
        str,
        ArgumentMetadata(
            display_name="Serialized JSON String",
            description="Deserialized JSON string.",
        ),
    ],
) -> JsonValue:
    return json.loads(serialized_json)


@action(
    display_name="Serialize JSON String",
    display_namespace="Admyral",
    description="Serializes a JSON string.",
)
def serialize_json_string(
    json_value: Annotated[
        JsonValue,
        ArgumentMetadata(
            display_name="JSON",
            description="A JSON value.",
        ),
    ],
) -> str:
    return json.dumps(json_value)


@action(
    display_name="Wait",
    display_namespace="Admyral",
    description="Waits for a specified number of seconds.",
)
def wait(
    seconds: Annotated[
        int,
        ArgumentMetadata(
            display_name="Seconds",
            description="Number of seconds to wait.",
        ),
    ],
) -> None:
    time.sleep(seconds)


@action(
    display_name="Transform",
    display_namespace="Admyral",
    description="Transforms a JSON.",
)
def transform(
    value: Annotated[
        JsonValue,
        ArgumentMetadata(
            display_name="Value",
            description="A JSON value.",
        ),
    ],
) -> JsonValue:
    return value



from temporalio import activity
from uuid import uuid4
import time

from admyral.workers.shared_worker_state import SharedWorkerState
from admyral.logger import get_logger


logger = get_logger(__name__)


@activity.defn
async def store_reference_resolution_error(
    run_id: str,
    prev_step_id: str,
    action_type: str,
    error: str,
) -> str:
    start = time.monotonic_ns()
    step_id = str(uuid4())
    await SharedWorkerState.get_store().store_workflow_run_error(
        step_id,
        run_id,
        action_type,
        prev_step_id,
        error,
        input_args={},
    )
    end = time.monotonic_ns()
    logger.info(f"Stored reference resolution error in {(end - start) / 1_000_000}ms")



from typing import Annotated
from httpx import Client
from datetime import timedelta

from admyral.action import action, ArgumentMetadata
from admyral.context import ctx
from admyral.typings import JsonValue
from admyral.utils.time import utc_now


def _fetch_token(
    client_id: str,
    client_secret: str,
    auth_url: str,
) -> str:
    with Client(
        headers={
            "Content-Type": "application/json",
            "Accept": "application/json",
        }
    ) as client:
        response = client.post(
            auth_url,
            json={
                "client_id": client_id,
                "client_secret": client_secret,
                "audience": "wiz-api",
                "grant_type": "client_credentials",
            },
        )
        response.raise_for_status()
        return response.json()["access_token"]


def get_wiz_client(
    client_id: str,
    client_secret: str,
    auth_url: str,
    api_endpoint: str,
) -> Client:
    access_token = _fetch_token(client_id, client_secret, auth_url)
    return Client(
        base_url=api_endpoint,
        headers={
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Authorization": f"Bearer {access_token}",
        },
    )


_LIST_ALERTS_QUERY = """
query IssuesTable($filterBy: IssueFilters, $first: Int, $after: String, $orderBy: IssueOrder) {
    issues: issuesV2(filterBy: $filterBy, first: $first, after: $after, orderBy: $orderBy) {
        nodes {
            id
            control {
                id
                name
                description
                resolutionRecommendation
                securitySubCategories {
                    title
                    category {
                        name
                        framework {
                            name
                        }
                    }
                }
            }
            createdAt
            updatedAt
            sourceRule {
                id
                name
            }
            dueAt
            resolvedAt
            statusChangedAt
            project {
                id
                name
                slug
                businessUnit
                riskProfile {
                    businessImpact
                }
            }
            status
            severity
            type
            entitySnapshot {
                id
                type
                nativeType
                name
                status
                cloudPlatform
                cloudProviderURL
                providerId
                region
                resourceGroupExternalId
                subscriptionExternalId
                subscriptionName
                subscriptionTags
                tags
                externalId
            }
            notes {
                createdAt
                updatedAt
                text
                user {
                    name
                    email
                }
                serviceAccount {
                    name
                }
            }
            serviceTickets {
                externalId
                name
                url
            }
        }
        pageInfo {
            hasNextPage
            endCursor
        }
    }
}
"""


# TODO: OCSF schema
@action(
    display_name="List Alerts",
    display_namespace="Wiz",
    description="List alerts from Wiz. If no time range is provided, the alerts from the last 24 hours will be listed.",
    secrets_placeholders=["WIZ_SECRET"],
)
def list_wiz_alerts(
    start_time: Annotated[
        str | None,
        ArgumentMetadata(
            display_name="Start Time",
            description="The start time for the cases to list. Must be in ISO 8601 format (YYYY-MM-DDTHH:MM:SSZ).",
        ),
    ] = None,
    end_time: Annotated[
        str | None,
        ArgumentMetadata(
            display_name="End Time",
            description="The end time for the cases to list. Must be in ISO 8601 format (YYYY-MM-DDTHH:MM:SSZ).",
        ),
    ] = None,
    limit: Annotated[
        int,
        ArgumentMetadata(
            display_name="Limit",
            description="The maximum number of cases to list.",
        ),
    ] = 1000,
) -> list[dict[str, JsonValue]]:
    # https://github.com/criblio/collector-templates/blob/main/collectors/rest/wiz/collector-wiz-issues.json

    secret = ctx.get().secrets.get("WIZ_SECRET")
    client_id = secret["client_id"]
    client_secret = secret["client_secret"]
    auth_url = secret["auth_url"]
    api_endpoint = secret["api_endpoint"]

    with get_wiz_client(client_id, client_secret, auth_url, api_endpoint) as client:
        alerts = []

        end_cursor = None
        variables = {
            "filterBy": {
                "sourceRule": {"id": []},
                "relatedEntity": {"type": []},
                "status": ["OPEN", "IN_PROGRESS", "RESOLVED", "REJECTED"],
                "severity": ["INFORMATIONAL", "LOW", "MEDIUM", "HIGH", "CRITICAL"],
                "type": [
                    "TOXIC_COMBINATION",
                    "THREAT_DETECTION",
                    "CLOUD_CONFIGURATION",
                ],
                "createdAt": {
                    "after": start_time
                    if start_time
                    else (utc_now() - timedelta(days=1)).strftime("%Y-%m-%dT%H:%M:%SZ"),
                    "before": end_time
                    if end_time
                    else utc_now().strftime("%Y-%m-%dT%H:%M:%SZ"),
                },
            },
            "first": 100,
        }

        while len(alerts) < limit:
            response = client.post(
                "",
                json={
                    "query": _LIST_ALERTS_QUERY,
                    "variables": variables,
                },
            )
            response.raise_for_status()
            data = response.json()["data"]["issues"]

            alerts.extend(data.get("nodes", []))

            if not data["pageInfo"]["hasNextPage"]:
                break

            if end_cursor := data["pageInfo"].get("endCursor"):
                variables["after"] = end_cursor
            else:
                break

        return alerts[:limit]



from enum import Enum
from pydantic import JsonValue  # noqa F401


class ScheduleType(str, Enum):
    CRON = "cron"
    INTERVAL_SECONDS = "interval_seconds"
    INTERVAL_MINUTES = "interval_minutes"
    INTERVAL_HOURS = "interval_hours"
    INTERVAL_DAYS = "interval_days"



from typing import TypeVar, Any, Callable

from admyral.utils.singleton import Singleton
from admyral.action import Action


F = TypeVar("F", bound=Callable[..., Any])


_RESERVED_ACTION_TYPES = {"start", "end", "if_condition", "note", "python"}


class ActionRegistry(metaclass=Singleton):
    _actions: dict[str, Action] = {}

    @classmethod
    def register(cls, type_name: str, action: Action) -> None:
        if type_name in cls._actions:
            raise ValueError(f"Action with type '{type_name}' already exists")
        if type_name in _RESERVED_ACTION_TYPES:
            raise ValueError(
                f"Action with type '{type_name}' is a reserved action. Please choose a different function name."
            )
        cls._actions[type_name] = action

    @classmethod
    def get_action_types(cls) -> set[str]:
        return set(cls._actions.keys())

    @classmethod
    def get_actions(cls) -> list[Action]:
        return list(cls._actions.values())

    @classmethod
    def get(cls, type_name: str) -> Action:
        if type_name not in cls._actions:
            raise ValueError(f"Action with type '{type_name}' does not exist")
        return cls._actions[type_name]

    @classmethod
    def is_registered(cls, type_name: str) -> bool:
        return type_name in cls._actions

    @classmethod
    def deregister(cls, type_name: str) -> None:
        if type_name in cls._actions:
            del cls._actions[type_name]


# needs to be defined after ActionRegistry, such that ActionRegistry is fully defined
# before any action is registered
from admyral.actions import *  # noqa: E402,F403



from temporalio.client import (
    Client,
    Schedule,
    ScheduleActionStartWorkflow,
    ScheduleSpec,
    ScheduleIntervalSpec,
)
from temporalio.common import RetryPolicy
from uuid import uuid4
from datetime import timedelta
import temporalio

from admyral.logger import get_logger
from admyral.workers.workflow_executor import WorkflowExecutor
from admyral.db.store_interface import StoreInterface
from admyral.models import WorkflowSchedule, Workflow
from admyral.typings import JsonValue


logger = get_logger(__name__)


RETRY_POLICY = RetryPolicy(maximum_attempts=1, non_retryable_error_types=["Exception"])


class WorkersClient:
    def __init__(self, store: StoreInterface, client: Client) -> None:
        self.client = client
        self.store = store

    @classmethod
    async def connect(cls, store: StoreInterface, host: str) -> "WorkersClient":
        logger.info(f"Connecting to Temporal at host {host}...")
        client = await Client.connect(host)
        return cls(store, client)

    async def execute_workflow(
        self,
        workflow_id: str,
        source_name: str,
        payload: dict[str, JsonValue] = {},
        trigger_default_args: dict[str, JsonValue] = {},
    ) -> None:
        logger.info(f"Executing workflow {workflow_id} from source {source_name}.")

        workflow = await self.store.get_workflow_by_id(workflow_id)
        if not workflow:
            raise ValueError(f"Workflow with id {workflow_id} not found.")

        # TODO: should we unify the temporal_workflow_id across triggers?
        temmporal_workflow_id = str(uuid4())
        await self.client.execute_workflow(
            WorkflowExecutor.run,
            {
                "workflow": workflow,
                "source_name": source_name,
                "payload": payload,
                "trigger_default_args": trigger_default_args,
            },
            id=temmporal_workflow_id,
            task_queue="workflow-queue",
            retry_policy=RETRY_POLICY,
        )

    def _build_temporal_schedule_spec(self, schedule: WorkflowSchedule) -> ScheduleSpec:
        if schedule.cron:
            return ScheduleSpec(cron_expressions=[schedule.cron])

        interval = None
        if schedule.interval_seconds:
            interval = timedelta(seconds=schedule.interval_seconds)
        elif schedule.interval_minutes:
            interval = timedelta(minutes=schedule.interval_minutes)
        elif schedule.interval_hours:
            interval = timedelta(hours=schedule.interval_hours)
        elif schedule.interval_days:
            interval = timedelta(hours=schedule.interval_days * 24)
        else:
            raise ValueError(
                "Invalid schedule trigger config. Neither an interval nor a cron expression was provided."
            )

        return ScheduleSpec(intervals=[ScheduleIntervalSpec(every=interval)])

    async def schedule_workflow(
        self, workflow: Workflow, schedule: WorkflowSchedule
    ) -> None:
        # TODO: should we unify the temporal_workflow_id across triggers?
        temmporal_workflow_id = str(uuid4())
        await self.client.create_schedule(
            schedule.schedule_id,
            Schedule(
                action=ScheduleActionStartWorkflow(
                    WorkflowExecutor.run,
                    {
                        "workflow": workflow,
                        "source_name": "schedule",
                        "payload": {},
                        "trigger_default_args": schedule.default_args,
                    },
                    id=temmporal_workflow_id,
                    task_queue="workflow-queue",
                    retry_policy=RETRY_POLICY,
                ),
                spec=self._build_temporal_schedule_spec(schedule),
            ),
        )

    async def delete_schedule(self, schedule_id: str) -> None:
        try:
            schedule_handle = self.client.get_schedule_handle(schedule_id)
            await schedule_handle.delete()
        except temporalio.service.RPCError as e:
            logger.error(f"Schedule ID {schedule_id} does not exist. Error: {e}")



from typing import Optional, Any, AsyncGenerator
from sqlalchemy.ext.asyncio import create_async_engine
from sqlmodel import select, delete, insert, update, text
from datetime import datetime
from sqlmodel.ext.asyncio.session import AsyncSession
from sqlalchemy.orm import sessionmaker
from contextlib import asynccontextmanager
from uuid import uuid4
import json

from admyral.models import (
    PipLockfile,
    PythonAction,
    Workflow,
    WorkflowRun,
    WorkflowWebhook,
    EncryptedSecret,
    WorkflowMetadata,
    ActionMetadata,
    SecretMetadata,
    WorkflowRunMetadata,
    WorkflowRunStepMetadata,
    WorkflowRunStep,
)
from admyral.models.workflow_schedule import WorkflowSchedule
from admyral.db.store_interface import StoreInterface
from admyral.db.schemas import (
    PythonActionSchema,
    PipLockfileCacheSchema,
    WorkflowSchema,
    WorkflowRunSchema,
    WorkflowRunStepsSchema,
    WorkflowWebhookSchema,
    WorkflowScheduleSchema,
    SecretsSchema,
)
from admyral.db.alembic.database_manager import DatabaseManager
from admyral.config.config import GlobalConfig, DatabaseType
from admyral.logger import get_logger
from admyral.utils.time import utc_now
from admyral.utils.crypto import generate_hs256
from admyral.typings import JsonValue


logger = get_logger(__name__)


class AdmyralDatabaseSession:
    def __init__(self, session: AsyncSession, execution_options: dict = {}) -> None:
        self.session = session
        self.execution_options = execution_options

    @classmethod
    async def from_session(
        cls,
        session: AsyncSession,
        execution_options: dict = {},
        is_sqlite: bool = False,
    ) -> "AdmyralDatabaseSession":
        db = cls(session, execution_options)
        if is_sqlite:
            await db.exec(text("PRAGMA foreign_keys = ON"))
        return db

    async def exec(self, statement: Any) -> Any:
        return await self.session.exec(
            statement, execution_options=self.execution_options
        )

    async def commit(self) -> None:
        await self.session.commit()


class AdmyralStore(StoreInterface):
    def __init__(self, config: GlobalConfig) -> None:
        self.config = config

        self.engine = create_async_engine(
            self.config.database_url, echo=True, future=True, pool_pre_ping=True
        )
        self.async_session_maker = sessionmaker(
            self.engine, class_=AsyncSession, expire_on_commit=False
        )

        self.execution_options = {}

        self.performed_setup = False

    # TODO: pass down config
    @classmethod
    async def create_store(cls, skip_setup: bool = False) -> "AdmyralStore":
        store = cls(GlobalConfig())
        if not skip_setup:
            await store.setup()
        store.performed_setup = True

        return store

    @classmethod
    async def create_test_store(cls) -> "AdmyralStore":
        config = GlobalConfig()
        config.database_type = DatabaseType.SQLITE
        config.database_url = "sqlite+aiosqlite:///:memory:"

        store = cls(config)
        await store.setup()
        store.performed_setup = True

        return store

    ########################################################
    # Setup & Chore
    ########################################################

    async def setup(self):
        logger.info("Setting up Admyral store.")

        database_manager = DatabaseManager(self.engine, self.config)

        does_db_exist = await database_manager.database_exists()
        if not does_db_exist:
            logger.info("Creating database...")
            await database_manager.create_database()
            logger.info("Database created.")

        logger.info("Running migrations...")
        await database_manager.migrate_database()
        logger.info("Migrations complete.")

        logger.info("Setting up Admyral store complete.")

    ########################################################
    # Helpers
    ########################################################

    # TODO: incompatability issue?
    @asynccontextmanager
    async def _get_async_session(self) -> AsyncGenerator[AdmyralDatabaseSession, None]:
        assert self.performed_setup, "Store has not been set up yet."
        async with self.async_session_maker() as session:
            db = await AdmyralDatabaseSession.from_session(
                session,
                self.execution_options,
                self.config.database_type == DatabaseType.SQLITE,
            )
            yield db

    ########################################################
    # Python Action
    ########################################################

    async def list_actions(self) -> list[ActionMetadata]:
        # TODO: consider user id
        user_id = self.config.user_id

        async with self._get_async_session() as db:
            result = await db.exec(
                select(PythonActionSchema).where(PythonActionSchema.user_id == user_id)
            )
            return [action.to_metadata() for action in result.all()]

    async def _get_action(
        self, db: AdmyralDatabaseSession, user_id: str, action_type: str
    ) -> Optional[PythonActionSchema]:
        result = await db.exec(
            select(PythonActionSchema)
            .where(PythonActionSchema.action_type == action_type)
            .where(PythonActionSchema.user_id == user_id),
        )
        return result.one_or_none()

    async def get_action(self, action_type: str) -> Optional[PythonAction]:
        # TODO: consider user id
        user_id = self.config.user_id

        async with self._get_async_session() as db:
            action = await self._get_action(db, user_id, action_type)
            return action.to_model() if action is not None else None

    async def store_action(self, action: PythonAction) -> None:
        # TODO: consider user id
        user_id = self.config.user_id

        async with self._get_async_session() as db:
            # if the action type already exists, we update the code and requirements.
            python_action_entry = await self._get_action(
                db, user_id, action.action_type
            )

            secrets_placeholders = (
                ";".join(action.secrets_placeholders)
                if action.secrets_placeholders
                else None
            )
            requirements = (
                ";".join(action.requirements) if action.requirements else None
            )

            if python_action_entry:
                await db.exec(
                    update(PythonActionSchema)
                    .where(PythonActionSchema.action_type == action.action_type)
                    .where(PythonActionSchema.user_id == user_id)
                    .values(
                        import_statements=action.import_statements,
                        code=action.code,
                        display_name=action.display_name,
                        display_namespace=action.display_namespace,
                        description=action.description,
                        secrets_placeholders=secrets_placeholders,
                        requirements=requirements,
                        arguments=[arg.model_dump() for arg in action.arguments],
                        updated_at=utc_now(),
                    )
                )
            else:
                await db.exec(
                    insert(PythonActionSchema).values(
                        user_id=user_id,
                        action_type=action.action_type,
                        import_statements=action.import_statements,
                        code=action.code,
                        display_name=action.display_name,
                        display_namespace=action.display_namespace,
                        description=action.description,
                        secrets_placeholders=secrets_placeholders,
                        requirements=requirements,
                        arguments=[arg.model_dump() for arg in action.arguments],
                    )
                )

            await db.commit()

    ########################################################
    # Pip Lockfile Cache
    ########################################################

    async def _get_cached_pip_lockfile(
        self, db: AdmyralDatabaseSession, hash: str
    ) -> Optional[PipLockfileCacheSchema]:
        expiration_time = utc_now()
        result = await db.exec(
            select(PipLockfileCacheSchema)
            .where(PipLockfileCacheSchema.hash == hash)
            .where(PipLockfileCacheSchema.expiration_time > expiration_time)
        )
        return result.one_or_none()

    async def get_cached_pip_lockfile(self, hash: str) -> Optional[PipLockfile]:
        async with self._get_async_session() as db:
            pip_lockfile = await self._get_cached_pip_lockfile(db, hash)
            return pip_lockfile.to_model() if pip_lockfile else None

    async def cache_pip_lockfile(
        self,
        pip_lockfile: PipLockfile,
    ) -> None:
        async with self._get_async_session() as db:
            # check whether the lockfile is already cached. if yes, then we update
            # its expiration time. otherwise, we create a new cache entry.
            cached_pip_lockfile = await self._get_cached_pip_lockfile(
                db, pip_lockfile.hash
            )

            if cached_pip_lockfile:
                await db.exec(
                    update(PipLockfileCacheSchema)
                    .where(PipLockfileCacheSchema.hash == pip_lockfile.hash)
                    .values(
                        expiration_time=datetime.fromtimestamp(
                            pip_lockfile.expiration_time
                        ),
                        updated_at=utc_now(),
                    )
                )
            else:
                await db.exec(
                    insert(PipLockfileCacheSchema).values(
                        hash=pip_lockfile.hash,
                        lockfile=pip_lockfile.lockfile,
                        expiration_time=datetime.fromtimestamp(
                            pip_lockfile.expiration_time
                        ),
                    )
                )

            await db.commit()

    async def delete_expired_cached_pip_lockfile(self) -> None:
        async with self._get_async_session() as db:
            current_time = utc_now()
            await db.exec(
                delete(PipLockfileCacheSchema).where(
                    PipLockfileCacheSchema.expiration_time < current_time
                )
            )
            await db.commit()

    ########################################################
    # Workflows
    ########################################################

    async def list_workflows(self) -> list[WorkflowMetadata]:
        # TODO: consider user id
        user_id = self.config.user_id

        async with self._get_async_session() as db:
            # TODO: loading all workflows just for the metadata
            # is not very efficient
            result = await db.exec(
                select(WorkflowSchema).where(WorkflowSchema.user_id == user_id)
            )
            return [workflow.to_metadata() for workflow in result.all()]

    async def _get_workflow_by_name(
        self, db: AdmyralDatabaseSession, user_id: str, workflow_name: str
    ) -> Optional[WorkflowSchema]:
        result = await db.exec(
            select(WorkflowSchema)
            .where(WorkflowSchema.workflow_name == workflow_name)
            .where(WorkflowSchema.user_id == user_id)
        )
        return result.one_or_none()

    async def get_workflow_by_name(self, workflow_name: str) -> Optional[Workflow]:
        # TODO: consider user id
        user_id = self.config.user_id

        async with self._get_async_session() as db:
            wf = await self._get_workflow_by_name(db, user_id, workflow_name)
            return wf.to_model() if wf else None

    async def _get_workflow_by_id(
        self, db: AdmyralDatabaseSession, user_id: str, workflow_id: str
    ) -> Optional[WorkflowSchema]:
        result = await db.exec(
            select(WorkflowSchema)
            .where(WorkflowSchema.workflow_id == workflow_id)
            .where(WorkflowSchema.user_id == user_id)
        )
        return result.one_or_none()

    async def get_workflow_by_id(self, workflow_id: str) -> Optional[Workflow]:
        # TODO: consider user id
        user_id = self.config.user_id

        async with self._get_async_session() as db:
            wf = await self._get_workflow_by_id(db, user_id, workflow_id)
            return wf.to_model() if wf else None

    async def store_workflow(self, workflow: Workflow) -> None:
        # TODO: consider user_id
        user_id = self.config.user_id

        async with self._get_async_session() as db:
            # if the workflow already exists, we update the workflow dag.
            stored_workflow = await self._get_workflow_by_id(
                db, user_id, workflow.workflow_id
            )

            if stored_workflow:
                await db.exec(
                    update(WorkflowSchema)
                    .where(WorkflowSchema.user_id == user_id)
                    .where(WorkflowSchema.workflow_id == workflow.workflow_id)
                    .values(
                        workflow_name=workflow.workflow_name,  # consider workflow name changes
                        workflow_dag=workflow.workflow_dag.model_dump(),
                        is_active=workflow.is_active,
                    )
                )
            else:
                await db.exec(
                    insert(WorkflowSchema).values(
                        workflow_id=workflow.workflow_id,
                        user_id=user_id,
                        workflow_name=workflow.workflow_name,
                        workflow_dag=workflow.workflow_dag.model_dump(),
                        is_active=workflow.is_active,
                    )
                )

            await db.commit()

    async def set_workflow_active_state(
        self, workflow_id: str, is_active: bool
    ) -> None:
        # TODO: consider user_id
        user_id = self.config.user_id

        async with self._get_async_session() as db:
            stored_workflow = await self._get_workflow_by_id(db, user_id, workflow_id)
            if not stored_workflow:
                raise ValueError(f"Workflow {workflow_id} not found.")

            await db.exec(
                update(WorkflowSchema)
                .where(WorkflowSchema.user_id == user_id)
                .where(WorkflowSchema.workflow_id == workflow_id)
                .values(is_active=is_active)
            )
            await db.commit()

    async def remove_workflow(self, workflow_id: str) -> None:
        # TODO: consider user_id
        user_id = self.config.user_id

        async with self._get_async_session() as db:
            await db.exec(
                delete(WorkflowSchema)
                .where(WorkflowSchema.user_id == user_id)
                .where(WorkflowSchema.workflow_id == workflow_id)
            )
            await db.commit()

    ########################################################
    # Workflow Webhooks
    ########################################################

    async def store_workflow_webhook(self, workflow_id: str) -> WorkflowWebhook:
        webhook_id = str(uuid4())
        webhook_secret = generate_hs256(webhook_id)

        async with self._get_async_session() as db:
            await db.exec(
                insert(WorkflowWebhookSchema).values(
                    webhook_id=webhook_id,
                    workflow_id=workflow_id,
                    webhook_secret=webhook_secret,
                )
            )
            await db.commit()

        return WorkflowWebhook(
            webhook_id=webhook_id,
            workflow_id=workflow_id,
            webhook_secret=webhook_secret,
        )

    async def get_webhook_for_workflow(
        self, workflow_id: str
    ) -> Optional[WorkflowWebhook]:
        async with self._get_async_session() as db:
            result = await db.exec(
                select(WorkflowWebhookSchema).where(
                    WorkflowWebhookSchema.workflow_id == workflow_id
                )
            )
            webhooks = result.all()

            if len(webhooks) == 0:
                return None
            if len(webhooks) > 1:
                # This should not happen!
                raise ValueError(f"Multiple webhooks found for workflow {workflow_id}")

            return webhooks[0].to_model()

    async def get_webhook(self, webhook_id: str) -> Optional[WorkflowWebhook]:
        async with self._get_async_session() as db:
            result = await db.exec(
                select(WorkflowWebhookSchema).where(
                    WorkflowWebhookSchema.webhook_id == webhook_id
                )
            )
            webhook = result.one_or_none()
            return webhook.to_model() if webhook else None

    async def delete_webhook(self, webhook_id: str) -> None:
        async with self._get_async_session() as db:
            await db.exec(
                delete(WorkflowWebhookSchema).where(
                    WorkflowWebhookSchema.webhook_id == webhook_id
                )
            )
            await db.commit()

    ########################################################
    # Workflow Schedules
    ########################################################

    async def store_schedule(self, schedule: WorkflowSchedule) -> None:
        # TODO: consider user id
        user_id = self.config.user_id

        async with self._get_async_session() as db:
            await db.exec(
                insert(WorkflowScheduleSchema).values(
                    schedule_id=schedule.schedule_id,
                    workflow_id=schedule.workflow_id,
                    user_id=user_id,
                    cron=schedule.cron,
                    interval_seconds=schedule.interval_seconds,
                    interval_minutes=schedule.interval_minutes,
                    interval_hours=schedule.interval_hours,
                    interval_days=schedule.interval_days,
                    default_args=schedule.default_args,
                )
            )
            await db.commit()

    async def list_schedules_for_workflow(
        self, workflow_id: str
    ) -> list[WorkflowSchedule]:
        # TODO: consider user id
        user_id = self.config.user_id

        async with self._get_async_session() as db:
            result = await db.exec(
                select(WorkflowScheduleSchema)
                .where(WorkflowScheduleSchema.user_id == user_id)
                .where(WorkflowScheduleSchema.workflow_id == workflow_id)
            )
            schedules = result.all()
            return [schedule.to_model() for schedule in schedules]

    async def delete_schedule(self, schedule_id: str) -> None:
        async with self._get_async_session() as db:
            await db.exec(
                delete(WorkflowScheduleSchema).where(
                    WorkflowScheduleSchema.schedule_id == schedule_id
                )
            )
            await db.commit()

    ########################################################
    # Workflow Runs
    ########################################################

    async def list_workflow_runs(
        self, workflow_id: str, limit: int = 100
    ) -> list[WorkflowRunMetadata]:
        # TODO: consider user_id
        user_id = self.config.user_id

        async with self._get_async_session() as db:
            result = await db.exec(
                select(WorkflowRunSchema)
                .where(WorkflowRunSchema.user_id == user_id)
                .where(WorkflowRunSchema.workflow_id == workflow_id)
                .order_by(WorkflowRunSchema.created_at.desc())
                .limit(limit)
            )
            return [workflow_run.to_metadata() for workflow_run in result.all()]

    async def init_workflow_run(
        self,
        run_id: str,
        step_id: str,
        workflow_id: str,
        source_name: str,
        payload: dict[str, JsonValue],
    ) -> None:
        # TODO: consider user_id
        user_id = self.config.user_id

        async with self._get_async_session() as db:
            await db.exec(
                insert(WorkflowRunSchema).values(
                    run_id=run_id,
                    user_id=user_id,
                    workflow_id=workflow_id,
                    source_name=source_name,
                )
            )

            await db.exec(
                insert(WorkflowRunStepsSchema).values(
                    step_id=step_id, run_id=run_id, action_type="start", result=payload
                )
            )

            await db.commit()

    async def mark_workflow_run_as_completed(
        self, run_id: str, completed_at: datetime
    ) -> None:
        # TODO: consider user_id
        user_id = self.config.user_id

        async with self._get_async_session() as db:
            await db.exec(
                update(WorkflowRunSchema)
                .where(WorkflowRunSchema.run_id == run_id)
                .where(WorkflowRunSchema.user_id == user_id)
                .values(
                    completed_at=completed_at,
                )
            )
            await db.commit()

    async def _get_workflow_run(
        self, db: AdmyralDatabaseSession, user_id: str, workflow_id: str, run_id: str
    ) -> Optional[WorkflowRunSchema]:
        result = await db.exec(
            select(WorkflowRunSchema)
            .where(WorkflowRunSchema.user_id == user_id)
            .where(WorkflowRunSchema.workflow_id == workflow_id)
            .where(WorkflowRunSchema.run_id == run_id)
        )
        return result.one_or_none()

    async def get_workflow_run(
        self, workflow_id: str, run_id: str
    ) -> Optional[WorkflowRun]:
        # TODO: consider user_id
        user_id = self.config.user_id

        async with self._get_async_session() as db:
            workflow_run = await self._get_workflow_run(
                db, user_id, workflow_id, run_id
            )
            return workflow_run.to_model() if workflow_run else None

    async def list_workflow_run_steps(
        self, workflow_id: str, run_id: str
    ) -> list[WorkflowRunStepMetadata]:
        # TODO: consider user_id and workflow_id
        user_id = self.config.user_id  # noqa F841

        async with self._get_async_session() as db:
            result = await db.exec(
                select(
                    WorkflowRunStepsSchema.step_id,
                    WorkflowRunStepsSchema.action_type,
                    WorkflowRunStepsSchema.error,
                )
                .where(WorkflowRunStepsSchema.run_id == run_id)
                .order_by(WorkflowRunStepsSchema.created_at)
            )
            return [
                WorkflowRunStepMetadata.model_validate(
                    {"step_id": row[0], "action_type": row[1], "error": row[2]}
                )
                for row in result.all()
            ]

    async def _get_workflow_run_step(
        self,
        db: AdmyralDatabaseSession,
        step_id: str,
    ) -> Optional[WorkflowRunStepsSchema]:
        # TODO: consider user_id + check when loading step_id

        result = await db.exec(
            select(WorkflowRunStepsSchema).where(
                WorkflowRunStepsSchema.step_id == step_id
            )
        )
        return result.one_or_none()

    async def get_workflow_run_step(
        self, workflow_id: str, run_id: str, step_id: str
    ) -> Optional[WorkflowRunStep]:
        # TODO: consider user_id + check when loading step_id

        async with self._get_async_session() as db:
            workflow_run_step = await self._get_workflow_run_step(db, step_id)
            return workflow_run_step.to_model() if workflow_run_step else None

    async def append_logs(
        self,
        step_id: str,
        run_id: str,
        action_type: str,
        prev_step_id: str,
        lines: list[str],
    ) -> None:
        async with self._get_async_session() as db:
            workflow_run_step = await self._get_workflow_run_step(db, step_id)

            if not workflow_run_step:
                await db.exec(
                    insert(WorkflowRunStepsSchema).values(
                        step_id=step_id,
                        run_id=run_id,
                        action_type=action_type,
                        prev_step_id=prev_step_id,
                        logs="\n".join(lines),
                    )
                )
            else:
                await db.exec(
                    update(WorkflowRunStepsSchema)
                    .where(WorkflowRunStepsSchema.step_id == step_id)
                    .values(logs=workflow_run_step.logs + "\n" + "\n".join(lines))
                )
            await db.commit()

    async def store_action_result(
        self,
        step_id: str,
        run_id: str,
        action_type: str,
        prev_step_id: str,
        result: JsonValue,
        input_args: dict[str, JsonValue],
    ) -> None:
        async with self._get_async_session() as db:
            workflow_run_step = await self._get_workflow_run_step(db, step_id)

            if not workflow_run_step:
                await db.exec(
                    insert(WorkflowRunStepsSchema).values(
                        step_id=step_id,
                        run_id=run_id,
                        action_type=action_type,
                        prev_step_id=prev_step_id,
                        result=result,
                        input_args=input_args,
                    )
                )
            else:
                await db.exec(
                    update(WorkflowRunStepsSchema)
                    .where(WorkflowRunStepsSchema.step_id == step_id)
                    .values(
                        result=result,
                        input_args=input_args,
                    )
                )
            await db.commit()

    async def store_workflow_run_error(
        self,
        step_id: str,
        run_id: str,
        action_type: str,
        prev_step_id: str,
        error: str,
        input_args: dict[str, JsonValue],
    ) -> None:
        # TODO: consider user_id
        user_id = self.config.user_id

        async with self._get_async_session() as db:
            workflow_run_step = await self._get_workflow_run_step(db, step_id)

            if not workflow_run_step:
                await db.exec(
                    insert(WorkflowRunStepsSchema).values(
                        step_id=step_id,
                        run_id=run_id,
                        action_type=action_type,
                        prev_step_id=prev_step_id,
                        input_args=input_args,
                        error=error,
                    )
                )
            else:
                await db.exec(
                    update(WorkflowRunStepsSchema)
                    .where(WorkflowRunStepsSchema.step_id == step_id)
                    .values(
                        error=error,
                        input_args=input_args,
                    )
                )

            await db.exec(
                update(WorkflowRunSchema)
                .where(WorkflowRunSchema.run_id == run_id)
                .where(WorkflowRunSchema.user_id == user_id)
                .values(failed_at=utc_now())
            )

            await db.commit()

    ########################################################
    # Secrets
    ########################################################

    async def list_secrets(
        self,
        user_id: str,
    ) -> list[SecretMetadata]:
        async with self._get_async_session() as db:
            result = await db.exec(
                select(SecretsSchema).where(SecretsSchema.user_id == user_id)
            )
            return [secret.to_metadata() for secret in result.all()]

    async def _get_secret(
        self, db: AdmyralDatabaseSession, user_id: str, secret_id: str
    ) -> Optional[SecretsSchema]:
        result = await db.exec(
            select(SecretsSchema)
            .where(SecretsSchema.user_id == user_id)
            .where(SecretsSchema.secret_id == secret_id)
        )
        return result.one_or_none()

    async def get_secret(
        self, user_id: str, secret_id: str
    ) -> Optional[EncryptedSecret]:
        async with self._get_async_session() as db:
            secret = await self._get_secret(db, user_id, secret_id)
            return secret.to_model() if secret else None

    async def store_secret(
        self,
        user_id: str,
        secret_id: str,
        encrypted_secret: Optional[str],
        schema: list[str],
    ) -> None:
        async with self._get_async_session() as db:
            secret = await self._get_secret(db, user_id, secret_id)

            if secret:
                await db.exec(
                    update(SecretsSchema)
                    .where(SecretsSchema.user_id == user_id)
                    .where(SecretsSchema.secret_id == secret_id)
                    .values(
                        encrypted_secret=encrypted_secret,
                        schema_json_serialized=json.dumps(schema),
                        updated_at=utc_now(),
                    )
                )
            else:
                await db.exec(
                    insert(SecretsSchema).values(
                        user_id=user_id,
                        secret_id=secret_id,
                        encrypted_secret=encrypted_secret,
                        schema_json_serialized=json.dumps(schema),
                    )
                )

            await db.commit()

    async def delete_secret(self, user_id: str, secret_id: str) -> None:
        async with self._get_async_session() as db:
            await db.exec(
                delete(SecretsSchema)
                .where(SecretsSchema.user_id == user_id)
                .where(SecretsSchema.secret_id == secret_id)
            )
            await db.commit()



from admyral.editor.graph_conversion import (
    workflow_to_editor_workflow_graph,
    editor_workflow_graph_to_workflow,
)

__all__ = ["workflow_to_editor_workflow_graph", "editor_workflow_graph_to_workflow"]



from admyral.actions.integrations.communication.slack import (
    send_slack_message,
    lookup_slack_user_by_email,
    send_slack_message_to_user_by_email,
    batched_send_slack_message_to_user_by_email,
)

__all__ = [
    "send_slack_message",
    "lookup_slack_user_by_email",
    "send_slack_message_to_user_by_email",
    "batched_send_slack_message_to_user_by_email",
]



import os
from typing import Annotated
import resend

from admyral.typings import JsonValue
from admyral.action import action, ArgumentMetadata


resend.api_key = os.getenv("RESEND_API_KEY")


@action(
    display_name="Send Email",
    display_namespace="Admyral",
    description="Send an email",
)
def send_email(
    recipients: Annotated[
        str | list[str],
        ArgumentMetadata(
            display_name="Recipients",
            description="The email addresses of the recipients",
        ),
    ],
    sender_name: Annotated[
        str,
        ArgumentMetadata(
            display_name="Sender Name",
            description="The name of the sender",
        ),
    ],
    subject: Annotated[
        str,
        ArgumentMetadata(
            display_name="Subject",
            description="The subject of the email",
        ),
    ],
    body: Annotated[
        str,
        ArgumentMetadata(
            display_name="Body",
            description="The body of the email",
        ),
    ],
) -> JsonValue:
    RESEND_EMAIL = os.getenv("RESEND_EMAIL")

    if not resend.api_key:
        raise ValueError(
            "RESEND_API_KEY environment variable is not set. Please set it to use the send_email action."
        )
    if not RESEND_EMAIL:
        raise ValueError(
            "RESEND_EMAIL environment variable is not set. Please set it to use the send_email action."
        )

    body = {
        "from": f"{sender_name} <{RESEND_EMAIL}>",
        "to": [recipients] if isinstance(recipients, str) else recipients,
        "subject": subject,
        "text": body,
    }

    # TODO: error handling
    return resend.Emails.send(body)
